@@ -23,14 +23,18 @@ module Collects
     module EmptyIteratorHandling
         export just_throws, may_use_type_inference
         using ..TypeUtil
-        const err = ArgumentError("couldn't figure out an appropriate element type")
+        @noinline function throw_err_eltype(collection_type::DataType)
+             err = ArgumentError(lazy"couldn't figure out an appropriate element type for collection of type $collection_type")
+             throw(err)
+        end
         """
             just_throws(iterator)::Union{}
 
         Throw an `ArgumentError`.
         """
-        function just_throws(::Any)
-            throw(err)
+        function just_throws(iterator)
+            t = typeof(iterator)
+            throw_err_eltype(t)
         end
         if isdefined(Base, Symbol("@default_eltype"))
             macro default_eltype(itr)
@@ -107,31 +111,46 @@ module Collects
         Int(length(collection))::Int
     end
 
-    Base.@constprop :aggressive function push!!(coll::Set, elem)
+    Base.@constprop :aggressive function typejoin_typeof_eltype(; coll, elem)
+        elt = eltype(coll)
+        typ = typeof(elem)
+        typejoin(elt, typ)
+    end
+
+    Base.@constprop :aggressive function push(coll::Set, elem)
+        E = typejoin_typeof_eltype(; coll, elem)
+        ret = Set{E}(coll)
+        push!(ret, elem)
+    end
+
+    Base.@constprop :aggressive function push_vector_prefix(coll, prefix_length::Int, elem)
+        E = typejoin_typeof_eltype(; coll, elem)
+        prefix = @view coll[begin:(begin + (prefix_length - 1))]
+        ret = Vector{E}(undef, prefix_length + 1)
+        ret = copyto!(ret, prefix)
+        ret[end] = elem
+        ret
+    end
+
+    Base.@constprop :aggressive function push(coll::Vector, elem)
+        len = length_int(coll)
+        push_vector_prefix(coll, len, elem)
+    end
+
+    Base.@constprop :aggressive function push!!(coll::Union{Set, Vector}, elem)
         elt = eltype(coll)
         if elem isa elt
             push!(coll, elem)
         else
-            let E = typejoin(typeof(elem), elt)
-                ret = Set{E}((elem,))
-                ret = sizehint!(ret, length_int(coll))
-                union!(ret, coll)
-            end
+            push(coll, elem)
         end
     end
 
-    Base.@constprop :aggressive function push!!(coll::Vector, elem)
-        elt = eltype(coll)
-        if elem isa elt
-            push!(coll, elem)
-        else
-            let E = typejoin(typeof(elem), elt)
-                ret = Vector{E}(undef, length_int(coll) + 1)
-                ret = copyto!(ret, coll)
-                ret[end] = elem
-                ret
-            end
+    Base.@constprop :aggressive function append!!(x::Union{Set, Vector}, y)
+        for e ∈ y
+            x = push!!(x, e)
         end
+        x
     end
 
     Base.@constprop :aggressive function zeros_tuple(n::Int)
@@ -151,8 +170,11 @@ module Collects
     end
 
     Base.@constprop :aggressive function check_ndims_consistency_impl(output::Int, input::Int)
+        @noinline function mismatch_throw(output::Int, input::Int)
+            throw(DimensionMismatch(lazy"dimension count mismatch: can't collect $input dimensions into $output dimensions"))
+        end
         if (!isone(output)) && (output != input)
-            throw(DimensionMismatch("dimension count mismatch: can't collect $input dimensions into $output dimensions"))
+            @noinline mismatch_throw(output, input)
         end
         output
     end
@@ -161,9 +183,19 @@ module Collects
         check_ndims_consistency_impl(ndims, infer_ndims(collection))
     end
 
+    const IteratorHasLength = Union{Base.HasLength, Base.HasShape}
+
+    Base.@constprop :aggressive function iterator_has_length(iterator)
+        Base.IteratorSize(iterator) isa IteratorHasLength
+    end
+
+    @noinline function throw_err_empty_union()
+        throw(ArgumentError("`Union{}` is not a type of a collection"))
+    end
+
     # Prevent accidental type piracy in dependent packages.
     function (::Collect)(::Type{Union{}}, ::Any)
-        throw(ArgumentError("`Union{}` is not a type of a collection"))
+        throw_err_empty_union()
     end
 
     const optional_memory = (@isdefined Memory) ? (Memory,) : ()
@@ -179,25 +211,74 @@ module Collects
         end,
     }
 
-    const ConstructorUnionFineInvariant = Union{
-        Type{Set},
+    const ConstructorUnionFineInvariantWithEltype = Union{
         (Type{Set{T}} where {T}),
-        Type{Array},
         (Type{Array{T}} where {T}),
+        (Type{Array{T, N}} where {T, N}),
+        if optional_memory === ()
+            Union{}
+        else
+            Type{Memory{T}} where {T}
+        end,
+    }
+
+    const ConstructorUnionFineInvariantWithoutEltype = Union{
+        Type{Set},
+        Type{Array},
+        (Type{Array{T, N} where {T}} where {N}),
+        if optional_memory === ()
+            Union{}
+        else
+            Type{Memory}
+        end,
+    }
+
+    const ConstructorUnionFineInvariant = Union{
+        ConstructorUnionFineInvariantWithEltype,
+        ConstructorUnionFineInvariantWithoutEltype,
+    }
+
+    const ConstructorUnionFineInvariantWithNdims = Union{
         (Type{Array{T, N} where {T}} where {N}),
         (Type{Array{T, N}} where {T, N}),
         if optional_memory === ()
             Union{}
         else
-            let M = only(optional_memory)
-                Union{
-                    Type{M},
-                    (Type{M{T}} where {T}),
-                }
-            end
+            Union{
+                Type{Memory},
+                (Type{Memory{T}} where {T}),
+            }
         end,
     }
 
+    const ConstructorUnionFineInvariantWithoutNdims = Union{
+        Type{Array},
+        (Type{Array{T}} where {T}),
+    }
+
+    Base.@constprop :aggressive function type_predicate_exhaustive(type::Type; true_supertype::Type{<:Type}, false_supertype::Type{<:Type})
+        let r
+            if type isa true_supertype
+                r = true
+            elseif type isa false_supertype
+                r = false
+            end
+            r
+        end
+    end
+
+    Base.@constprop :aggressive function type_has_eltype(type::Type)
+        true_supertype = ConstructorUnionFineInvariantWithEltype
+        false_supertype = ConstructorUnionFineInvariantWithoutEltype
+        type_predicate_exhaustive(type; true_supertype, false_supertype)
+    end
+
+    Base.@constprop :aggressive function type_has_ndims(type::Type)
+        true_supertype = ConstructorUnionFineInvariantWithNdims
+        false_supertype = ConstructorUnionFineInvariantWithoutNdims
+        type_predicate_exhaustive(type; true_supertype, false_supertype)
+    end
+
     Base.@constprop :aggressive function collect_as_set_with_unknown_eltype(e::E, collection) where {E}
         iter = Iterators.peel(collection)
         if iter === nothing
@@ -205,10 +286,7 @@ module Collects
         else
             let (fir, rest) = iter
                 coll = Set((fir,))
-                for e ∈ rest
-                    coll = push!!(coll, e)
-                end
-                coll
+                append!!(coll, rest)
             end
         end
     end
@@ -222,17 +300,34 @@ module Collects
         ret
     end
 
-    Base.@constprop :aggressive function collect_as_set(e::E, ::Type{Set}, collection) where {E}
-        T = eltype(collection)
-        if TypeUtil.is_precise(T)
-            collect_as_set_with_known_eltype(T, collection)
-        else
-            collect_as_set_with_unknown_eltype(e, collection)
-        end
+    Base.@constprop :aggressive function collect_as_vector_with_initial(initial, initial_length::Int, next, rest)
+        vec = push_vector_prefix(initial, initial_length, next)
+        append!!(vec, rest)
     end
 
-    Base.@constprop :aggressive function collect_as_set(::Any, ::Type{Set{T}}, collection) where {T}
-        collect_as_set_with_known_eltype(T, collection)
+    Base.@constprop :aggressive function collect_as_vector_with_unknown_eltype_and_known_length(type::Type, first, rest, len::Int)
+        T = typeof(first)
+        vec = type{T}(undef, len)
+        i = 1
+        vec[i] = first
+        while true
+            iter = Iterators.peel(rest)
+            if iter === nothing
+                break
+            end
+            (next, rest) = iter
+            if next isa T
+                i += 1
+                vec[i] = next
+            else
+                return collect_as_vector_with_initial(vec, i, next, rest)
+            end
+        end
+        vec
+    end
+
+    Base.@constprop :aggressive function collect_as_vector_with_unknown_eltype_and_unknown_length(first, rest)
+        append!!([first], rest)
     end
 
     Base.@constprop :aggressive function collect_as_array_with_unknown_eltype(e::E, ndims::Int, collection) where {E}
@@ -249,9 +344,10 @@ module Collects
                     Array{e(collection), ndims}(undef, zeros_tuple(ndims))
                 else
                     let (fir, rest) = iter
-                        vec = [fir]
-                        for e ∈ rest
-                            vec = push!!(vec, e)
+                        vec = if iterator_has_length(collection)
+                            collect_as_vector_with_unknown_eltype_and_known_length(Vector, fir, rest, length_int(collection))
+                        else
+                            collect_as_vector_with_unknown_eltype_and_unknown_length(fir, rest)
                         end
                         if isone(ndims)
                             vec
@@ -287,7 +383,7 @@ module Collects
             end
         else
             let V = Vector{T}
-                vec = if Base.IteratorSize(collection) isa Union{Base.HasLength, Base.HasShape}
+                vec = if iterator_has_length(collection)
                     collect_as_vectorlike_with_known_eltype_and_length(V, collection)
                 else
                     collect_as_vector_with_known_eltype(V, collection)
@@ -301,42 +397,6 @@ module Collects
         end
     end
 
-    Base.@constprop :aggressive function collect_as_array_with_optional_eltype(e::E, ::Type{T}, N::Int, collection) where {E, T}
-        if TypeUtil.is_precise(T)
-            collect_as_array_with_known_eltype(T, N, collection)
-        else
-            collect_as_array_with_unknown_eltype(e, N, collection)
-        end
-    end
-
-    Base.@constprop :aggressive function collect_as_array(e::E, ::Type{Array}, collection) where {E}
-        T = eltype(collection)
-        N = infer_ndims(collection)
-        collect_as_array_with_optional_eltype(e, T, N, collection)
-    end
-
-    Base.@constprop :aggressive function collect_as_array(::Any, ::Type{Array{T}}, collection) where {T}
-        N = infer_ndims(collection)
-        collect_as_array_with_known_eltype(T, N, collection)
-    end
-
-    Base.@constprop :aggressive function collect_as_array(e::E, ::Type{Array{T, N} where {T}}, collection) where {E, N}
-        T = eltype(collection)
-        collect_as_array_with_optional_eltype(e, T, N, collection)
-    end
-
-    Base.@constprop :aggressive function collect_as_array(::Any, ::Type{Array{T, N}}, collection) where {T, N}
-        collect_as_array_with_known_eltype(T, N, collection)
-    end
-
-    Base.@constprop :aggressive function collect_as_common_invariant(e::E, ::Type{C}, collection) where {E, C <: Set}
-        collect_as_set(e, C, collection)
-    end
-
-    Base.@constprop :aggressive function collect_as_common_invariant(e::E, ::Type{C}, collection) where {E, C <: Array}
-        collect_as_array(e, C, collection)
-    end
-
     if optional_memory !== ()
         Base.@constprop :aggressive function collect_as_memory_with_known_eltype_and_known_length(::Type{T}, collection::Memory{T}) where {T}
             copy(collection)
@@ -349,31 +409,74 @@ module Collects
             collect_as_memory_with_known_eltype_and_known_length(T, vec)
         end
         Base.@constprop :aggressive function collect_as_memory_with_known_eltype(::Type{T}, collection) where {T}
-            if Base.IteratorSize(collection) isa Union{Base.HasLength, Base.HasShape}
+            if iterator_has_length(collection)
                 collect_as_memory_with_known_eltype_and_known_length(T, collection)
             else
                 collect_as_memory_with_known_eltype_and_unknown_length(T, collection)
             end
         end
-        Base.@constprop :aggressive function collect_as_memory_with_unknown_eltype(e::E, collection) where {E}
-            vec = collect_as_array_with_unknown_eltype(e, 1, collection)
-            collect_as_memory_with_known_eltype_and_known_length(eltype(vec), vec)
-        end
-        Base.@constprop :aggressive function collect_as_memory_with_optional_eltype(e::E, ::Type{T}, collection) where {E, T}
-            if TypeUtil.is_precise(T)
-                collect_as_memory_with_known_eltype(T, collection)
+        Base.@constprop :aggressive function collect_as_memory_with_unknown_eltype_and_known_length(e::E, collection) where {E}
+            iter = Iterators.peel(collection)
+            if iter === nothing
+                return Memory{e(collection)}(undef, 0)
+            end
+            (fir, rest) = iter
+            vec = collect_as_vector_with_unknown_eltype_and_known_length(Memory, fir, rest, length_int(collection))
+            if vec isa Memory
+                vec
             else
-                collect_as_memory_with_unknown_eltype(e, collection)
+                collect_as_memory_with_known_eltype(eltype(vec), vec)
             end
         end
-        Base.@constprop :aggressive function collect_as_memory(e::E, ::Type{only(optional_memory)}, collection) where {E}
-            collect_as_memory_with_optional_eltype(e, eltype(collection), collection)
+        Base.@constprop :aggressive function collect_as_memory_with_unknown_eltype_and_unknown_length(e::E, collection) where {E}
+            vec = collect_as_array_with_unknown_eltype(e, 1, collection)
+            collect_as_memory_with_known_eltype(eltype(vec), vec)
         end
-        Base.@constprop :aggressive function collect_as_memory(::Any, ::Type{only(optional_memory){T}}, collection) where {T}
-            collect_as_memory_with_known_eltype(T, collection)
+        Base.@constprop :aggressive function collect_as_memory_with_unknown_eltype(e::E, collection) where {E}
+            if iterator_has_length(collection)
+                collect_as_memory_with_unknown_eltype_and_known_length(e, collection)
+            else
+                collect_as_memory_with_unknown_eltype_and_unknown_length(e, collection)
+            end
         end
-        Base.@constprop :aggressive function collect_as_common_invariant(e::E, ::Type{C}, collection) where {E, C <: only(optional_memory)}
-            collect_as_memory(e, C, collection)
+    end
+
+    Base.@constprop :aggressive function collect_as_common_invariant(e::E, type::Type, collection) where {E}
+        has_eltype = type_has_eltype(type)
+        elt = if has_eltype
+            eltype(type)
+        else
+            eltype(collection)
+        end
+        subtypes_set = type <: Set
+        subtypes_memory = (optional_memory !== ()) && (type <: Memory)
+        subtypes_array = type <: Array
+        let r, n
+            if subtypes_array
+                n = if type_has_ndims(type)
+                    ndims(type)::Int
+                else
+                    infer_ndims(collection)::Int
+                end
+            end
+            if has_eltype || TypeUtil.is_precise(elt)
+                if subtypes_set
+                    r = collect_as_set_with_known_eltype(elt, collection)
+                elseif subtypes_memory
+                    r = collect_as_memory_with_known_eltype(elt, collection)
+                elseif subtypes_array
+                    r = collect_as_array_with_known_eltype(elt, n, collection)
+                end
+            else
+                if subtypes_set
+                    r = collect_as_set_with_unknown_eltype(e, collection)
+                elseif subtypes_memory
+                    r = collect_as_memory_with_unknown_eltype(e, collection)
+                elseif subtypes_array
+                    r = collect_as_array_with_unknown_eltype(e, n, collection)
+                end
+            end
+            r
         end
     end
 
@@ -405,9 +508,12 @@ module Collects
     end
 
     Base.@constprop :aggressive function (collect::Collect)(type::ConstructorUnionRough, collection)
-        if Base.IteratorSize(collection) === Base.IsInfinite()
+        @noinline function infinite_throw()
             throw(ArgumentError("can't collect infinitely many elements into a finite collection"))
         end
+        if Base.IteratorSize(collection) === Base.IsInfinite()
+            @noinline infinite_throw()
+        end
         collect_as_common(collect.empty_iterator_handler, TypeUtil.normalize(type), collection)
     end
 
@@ -424,4 +530,23 @@ module Collects
         c = Collect(; empty_iterator_handler)
         c(T, collection)
     end
+
+    """
+        collect_as(output_type::Type)
+
+    Return a callable which:
+
+    * Takes a collection, `collection`, as the only positional argument.
+
+    * Takes the same keyword arguments as the `collect_as` method with two positional arguments. Say, `kwargs`.
+
+    * Calls `collect_as(output_type, collection; kwargs...)` and returns the result.
+
+    Mostly equivalent to `Base.Fix1(collect_as, output_type)`.
+    """
+    Base.@constprop :aggressive function collect_as(::Type{T}) where {T}
+        Base.@constprop :aggressive function collect_as_with_fixed_requested_output_type(collection; empty_iterator_handler::EIH = just_throws) where {EIH}
+            collect_as(T, collection; empty_iterator_handler)
+        end
+    end
 end
